@@ -1,4 +1,4 @@
-/* Copyright (c) 1993-2000
+/* Copyright (c) 1993-2002
  *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
  *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
  * Copyright (c) 1987 Oliver Laumann
@@ -54,6 +54,18 @@ struct aclusergroup
  *  ==> user.h
  */
 
+/*
+ * a copy buffer
+ */
+struct plop
+{
+  char *buf;
+  int len;
+#ifdef ENCODINGS
+  int enc;
+#endif
+};
+
 /*
  * A User has a list of groups, and points to other users.  
  * users is the User entry of the session owner (creator)
@@ -69,8 +81,7 @@ typedef struct acluser
   int  u_detachotherwin;	/* window that was "other" when he detached */
   int  u_Esc, u_MetaEsc;	/* the users screen escape character */
 #ifdef COPY_PASTE
-  char  *u_copybuffer;
-  int   u_copylen;
+  struct plop u_plop;
 #endif
 #ifdef MULTIUSER
   int u_id;			/* a uniq index in the bitfields. */
