@@ -1,4 +1,4 @@
-/* Copyright (c) 1993-2000
+/* Copyright (c) 1993-2002
  *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
  *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
  * Copyright (c) 1987 Oliver Laumann
@@ -216,7 +216,6 @@ sched()
        * Sequents select emulation counts a descriptor which is
        * readable and writeable only as one hit. Waaaaa.
        */
-      debug2("nsel %d, was %d; ", nsel ? 2 * FD_SETSIZE : nsel, nsel);
       if (nsel)
         nsel = 2 * FD_SETSIZE;
 #endif
@@ -279,7 +278,7 @@ static int sgihack()
 	{
 	  if (errno == EINTR)
 	    continue;
-	  SigHup(SIGARG);	/* goodbye display */
+	  Hangup();	/* goodbye display */
 	  return 1;
 	}
       display = display->d_next;
