@@ -1,4 +1,4 @@
-/* Copyright (c) 1993-2001
+/* Copyright (c) 1993-2002
  *      Juergen Weigert (jnweiger@immd4.informatik.uni-erlangen.de)
  *      Michael Schroeder (mlschroe@immd4.informatik.uni-erlangen.de)
  * Copyright (c) 1987 Oliver Laumann
@@ -75,8 +75,10 @@ RCS_ID("$Id: screen.c,v 1.24 1994/09/06 17:00:20 mlschroe Exp $ FAU")
 #if (defined(AUX) || defined(_AUX_SOURCE)) && defined(POSIX)
 # include <compat.h>
 #endif
-#if defined(HAVE_NL_LANGINFO) && defined(UTF8)
+#if defined(USE_LOCALE) || defined(ENCODINGS)
 # include <locale.h>
+#endif
+#if defined(HAVE_NL_LANGINFO) && defined(ENCODINGS)
 # include <langinfo.h>
 #endif
 
@@ -108,16 +110,13 @@ FILE *dfp;
 #endif
 
 
-extern char *blank, *null, Term[], screenterm[], **environ, Termcap[];
+extern char Term[], screenterm[], **environ, Termcap[];
 int force_vt = 1;
 int VBellWait, MsgWait, MsgMinWait, SilenceWait;
 
 extern struct acluser *users;
 extern struct display *displays, *display; 
 
-/* tty.c */
-extern int intrc;
-
 
 extern int visual_bell;
 #ifdef COPY_PASTE
@@ -186,6 +185,9 @@ char *PowDetachString;
 #endif
 char *hstatusstring;
 char *captionstring;
+char *timestring;
+char *wliststr;
+char *wlisttit;
 int auto_detach = 1;
 int iflag, rflag, dflag, lsflag, quietflag, wipeflag, xflag;
 int cmdflag;
@@ -208,6 +210,10 @@ int default_startup;
 int ZombieKey_destroy, ZombieKey_resurrect;
 char *preselect = NULL;		/* only used in Attach() */
 
+#ifdef UTF8
+char *screenencodings;
+#endif
+
 #ifdef NETHACK
 int nethackflag = 0;
 #endif
@@ -271,7 +277,7 @@ struct passwd *ppp;
   static char *spw = NULL;
 #endif
  
-  if (!(ppp = getpwnam(name)))
+  if (!ppp && !(ppp = getpwnam(name)))
     return NULL;
 
   /* Do password sanity check..., allow ##user for SUN_C2 security */
@@ -426,6 +432,9 @@ char **av;
   logtstamp_string = SaveStr("-- %n:%t -- time-stamp -- %M/%d/%y %c:%s --\n");
   hstatusstring = SaveStr("%h");
   captionstring = SaveStr("%3n %t");
+  timestring = SaveStr("%c:%s %M %d %H%? %l%?");
+  wlisttit = SaveStr("Num Name%=Flags");
+  wliststr = SaveStr("%3n %t%=%f");
 #ifdef COPY_PASTE
   BufferFile = SaveStr(DEFAULT_BUFFERFILE);
 #endif
@@ -445,6 +454,10 @@ char **av;
 
 #ifdef COPY_PASTE
   CompileKeys((char *)NULL, mark_key_tab);
+#endif
+#ifdef UTF8
+  InitBuiltinTabs();
+  screenencodings = SaveStr(SCREENENCODINGS);
 #endif
   nwin = nwin_undef;
   nwin_options = nwin_undef;
@@ -579,6 +592,9 @@ char **av;
 		    case '1':
 		      nwin_options.lflag = 1;
 		      break;
+		    case 'a':
+		      nwin_options.lflag = 3;
+		      break;
 		    case 's':	/* -ls */
 		    case 'i':	/* -list */
 		      lsflag = 1;
@@ -686,7 +702,7 @@ char **av;
 		  /* NOTREACHED */
 #ifdef UTF8
 		case 'U':
-		  nwin_options.utf8 = nwin_options.utf8 == -1 ? 1 : 0;
+		  nwin_options.encoding = nwin_options.encoding == -1 ? UTF8 : 0;
 		  break;
 #endif
 		default:
@@ -697,23 +713,27 @@ char **av;
       else
 	break;
     }
-#ifdef UTF8
-  if (nwin_options.utf8 == -1)
+#ifdef USE_LOCALE
+  setlocale(LC_ALL, "");
+#endif
+#ifdef ENCODINGS
+  if (nwin_options.encoding == -1)
     {
       /* ask locale if we should start in UTF-8 mode */
 # ifdef HAVE_NL_LANGINFO
-      debug("asking locale for CODESET\n");
+#  ifndef USE_LOCALE
       setlocale(LC_CTYPE, "");
-      if (strcmp(nl_langinfo(CODESET), "UTF-8") == 0)
-        nwin_options.utf8 = 1;
-      debug1("utf8 = %d\n", nwin_options.utf8);
+#  endif
+      nwin_options.encoding = FindEncoding(nl_langinfo(CODESET));
+      debug1("locale says encoding = %d\n", nwin_options.encoding);
 # else
+#  ifdef UTF8
       char *s;
-      debug("asking environment for LC_CTYPE\n");
       if (((s = getenv("LC_ALL")) || (s = getenv("LC_CTYPE")) ||
           (s = getenv("LANG"))) && InStr(s, "UTF-8"))
-        nwin_options.utf8 = 1;
-      debug1("utf=%d\n", nwin_options.utf8);
+        nwin_options.encoding = UTF8;
+#  endif
+      debug1("environment says encoding=%d\n", nwin_options.encoding);
 #endif
     }
 #endif
@@ -724,6 +744,9 @@ char **av;
   if (!cmdflag && dflag && mflag && !(rflag || xflag))
     detached = 1;
   nwin = nwin_options;
+#ifdef ENCODINGS
+  nwin.encoding = nwin_undef.encoding;	/* let screenrc overwrite it */
+#endif
   if (ac)
     nwin.args = av;
   real_uid = getuid();
@@ -939,7 +962,7 @@ char **av;
 	  SockDir = SOCKDIR;
 	  if (lstat(SockDir, &st))
 	    {
-	      n = (eff_uid == 0) ? 0755 :
+	      n = (eff_uid == 0 && (real_uid || eff_gid == real_gid)) ? 0755 :
 	          (eff_gid != real_gid) ? 0775 :
 #ifdef S_ISVTX
 		  0777|S_ISVTX;
@@ -1186,8 +1209,9 @@ char **av;
     {
       if (MakeDisplay(LoginName, attach_tty, attach_term, n, getppid(), &attach_Mode) == 0)
 	Panic(0, "Could not alloc display");
-#ifdef UTF8
-      D_utf8 = nwin_options.utf8 == 1;
+#ifdef ENCODINGS
+      D_encoding = nwin_options.encoding > 0 ? nwin_options.encoding : 0;
+      debug1("D_encoding = %d\n", D_encoding);
 #endif
     }
 
@@ -1264,7 +1288,7 @@ char **av;
   if (display)
     {
       brktty(D_userfd);
-      SetMode(&D_OldMode, &D_NewMode, display->d_flow, iflag);
+      SetMode(&D_OldMode, &D_NewMode, D_flow, iflag);
       /* Note: SetMode must be called _before_ FinishRc. */
       SetTTY(D_userfd, &D_NewMode);
       if (fcntl(D_userfd, F_SETFL, FNBLOCK))
@@ -1357,6 +1381,7 @@ struct win *p;
       p->w_y = MFindUsedLine(p, p->w_bot, 1);
       sprintf(buf, "\n\r=== Window terminated (%s) ===", s ? s : "?");
       WriteString(p, buf, strlen(buf));
+      WindowChanged(p, 'f');
     }
   else
     KillWindow(p);
@@ -1406,18 +1431,9 @@ SigChld SIGDEFARG
 sigret_t
 SigHup SIGDEFARG
 {
-  if (display == 0)
-    return;
-  debug("SigHup()\n");
-  if (D_userfd >= 0)
-    {
-      close(D_userfd);
-      D_userfd = -1;
-    }
-  if (auto_detach || displays->d_next)
-    Detach(D_HANGUP);
-  else
-    Finit(0);
+  /* Hangup all displays */
+  while ((display = displays) != 0)
+    Hangup();
   SIGRETURN;
 }
 
@@ -1430,14 +1446,19 @@ static sigret_t
 SigInt SIGDEFARG
 {
 #if HAZARDOUS
-  char buf[1];
+  char ibuf;
 
   debug("SigInt()\n");
-  *buf = (char) intrc;
-  if (fore)
-    fore->w_inlen = 0;
-  if (fore)
-    write(fore->w_ptyfd, buf, 1);
+  if (fore && displays)
+    {
+# if defined(TERMIO) || defined(POSIX)
+      ibuf = displays->d_OldMode.tio.c_cc[VINTR];
+# else
+      ibuf = displays->d_OldMode.m_tchars.t_intrc;
+# endif
+      fore->w_inlen = 0;
+      write(fore->w_ptyfd, &ibuf, 1);
+    }
 #else
   signal(SIGINT, SigInt);
   debug("SigInt() careful\n");
@@ -1646,6 +1667,23 @@ int e;
   exit(e);
 }
 
+void
+Hangup()
+{
+  if (display == 0)
+    return;
+  debug1("Hangup %x\n", display);
+  if (D_userfd >= 0)
+    {
+      close(D_userfd);
+      D_userfd = -1;
+    }
+  if (auto_detach || displays->d_next)
+    Detach(D_HANGUP);
+  else
+    Finit(0);
+}
+
 /*
  * Detach now has the following modes:
  *D_DETACH	 SIG_BYE	detach backend and exit attacher
@@ -1675,6 +1713,8 @@ int mode;
   if (D_status)
     RemoveStatus();
   FinitTerm();
+  if (!display)
+    return;
   switch (mode)
     {
     case D_HANGUP:
@@ -1728,7 +1768,7 @@ int mode;
     {
       for (p = windows; p; p = p->w_next)
         {
-	  if (p->w_slot != (slot_t) -1)
+	  if (p->w_slot != (slot_t) -1 && !(p->w_lflag & 2))
 	    {
 	      RemoveUtmp(p);
 	      /*
@@ -1967,8 +2007,10 @@ VA_DECL
  *
  */
 
+#ifndef USE_LOCALE
 static const char days[]   = "SunMonTueWedThuFriSat";
 static const char months[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
+#endif
 
 static char winmsg_buf[MAXSTR];
 #define MAX_WINMSG_REND 16	/* rendition changes */
@@ -1976,11 +2018,48 @@ static int winmsg_rend[MAX_WINMSG_REND];
 static int winmsg_rendpos[MAX_WINMSG_REND];
 static int winmsg_numrend;
 
+static char *
+pad_expand(buf, p, numpad, padlen)
+char *buf;
+char *p;
+int numpad;
+int padlen;
+{
+  char *pn, *pn2;
+  int i, r;
+
+  padlen = padlen - (p - buf);	/* space for rent */
+  if (padlen < 0)
+    padlen = 0;
+  pn2 = pn = p + padlen;
+  r = winmsg_numrend;
+  while (p >= buf)
+    {
+      if (r && p - buf == winmsg_rendpos[r - 1])
+	{
+	  winmsg_rendpos[--r] = pn - buf;
+	  continue;
+	}
+      *pn-- = *p;
+      if (*p-- == 127)
+	{
+	  pn[1] = ' ';
+	  i = numpad > 0 ? (padlen + numpad - 1) / numpad : 0;
+	  padlen -= i;
+	  while (i-- > 0)
+	    *pn-- = ' ';
+	  numpad--;
+	}
+    }
+  return pn2;
+}
+
 char *
-MakeWinMsgEv(str, win, esc, ev)
+MakeWinMsgEv(str, win, esc, padlen, ev)
 char *str;
 struct win *win;
 int esc;
+int padlen;
 struct event *ev;
 {
   static int tick;
@@ -1992,8 +2071,16 @@ struct event *ev;
   int l, i, r;
   int num;
   int zeroflg;
-  int qmflag = 0, omflag = 0;
+  int longflg;
+  int minusflg;
+  int plusflg;
+  int qmflag = 0, omflag = 0, qmnumrend = 0;
   char *qmpos = 0;
+  int numpad = 0;
+  int lastpad = 0;
+  int truncpos = -1;
+  int truncper = 0;
+  int trunclong = 0;
  
   if (winmsg_numrend >= 0)
     winmsg_numrend = 0;
@@ -2037,11 +2124,17 @@ struct event *ev;
 	}
       if (*++s == esc)	/* double escape ? */
 	continue;
+      if ((plusflg = *s == '+') != 0)
+	s++;
+      if ((minusflg = *s == '-') != 0)
+	s++;
       if ((zeroflg = *s == '0') != 0)
 	s++;
       num = 0;
       while(*s >= '0' && *s <= '9')
 	num = num * 10 + (*s++ - '0');
+      if ((longflg = *s == 'L') != 0)
+	s++;
       switch (*s)
 	{
         case '?':
@@ -2049,11 +2142,16 @@ struct event *ev;
 	  if (qmpos)
 	    {
 	      if ((!qmflag && !omflag) || omflag == 1)
-	        p = qmpos;
+		{
+	          p = qmpos;
+	          if (qmnumrend < winmsg_numrend)
+		    winmsg_numrend = qmnumrend;
+		}
 	      qmpos = 0;
 	      break;
 	    }
 	  qmpos = p;
+	  qmnumrend = winmsg_numrend;
 	  qmflag = omflag = 0;
 	  break;
         case ':':
@@ -2064,10 +2162,13 @@ struct event *ev;
 	    {
 	      omflag = 1;
 	      qmpos = p;
+	      qmnumrend = winmsg_numrend;
 	    }
 	  else
 	    {
 	      p = qmpos;
+	      if (qmnumrend < winmsg_numrend)
+		winmsg_numrend = qmnumrend;
 	      omflag = -1;
 	    }
 	  break;
@@ -2088,7 +2189,11 @@ struct event *ev;
 	      tick |= 4;
 	      break;
 	    case 'D':
+#ifdef USE_LOCALE
+	      strftime(p, l, (longflg ? "%A" : "%a"), tm);
+#else
 	      sprintf(p, "%3.3s", days + 3 * tm->tm_wday);
+#endif
 	      tick |= 4;
 	      break;
 	    case 'm':
@@ -2096,7 +2201,11 @@ struct event *ev;
 	      tick |= 4;
 	      break;
 	    case 'M':
+#ifdef USE_LOCALE
+	      strftime(p, l, (longflg ? "%B" : "%b"), tm);
+#else
 	      sprintf(p, "%3.3s", months + 3 * tm->tm_mon);
+#endif
 	      tick |= 4;
 	      break;
 	    case 'y':
@@ -2178,12 +2287,16 @@ struct event *ev;
 	case 'W':
 	  {
 	    struct win *oldfore = 0;
+	    char *ss;
+
 	    if (display)
 	      {
 		oldfore = D_fore;
 		D_fore = win;
 	      }
-	    AddWindows(p, l - 1, *s == 'w' ? 2 : 3, -1);
+	    ss = AddWindows(p, l - 1, (*s == 'w' ? 0 : 1) | (longflg ? 0 : 2) | (plusflg ? 4 : 0), win ? win->w_number : -1);
+	    if (minusflg)
+	       *ss = 0;
 	    if (display)
 	      D_fore = oldfore;
 	  }
@@ -2199,6 +2312,14 @@ struct event *ev;
 	    qmflag = 1;
 	  p += strlen(p) - 1;
 	  break;
+	case 'f':
+	  *p = 0;
+	  if (win)
+	    AddWindowFlags(p, l - 1, win);
+	  if (*p)
+	    qmflag = 1;
+	  p += strlen(p) - 1;
+	  break;
 	case 't':
 	  *p = 0;
 	  if (win && strlen(win->w_title) < l)
@@ -2236,6 +2357,141 @@ struct event *ev;
 	    p--;
           }
 	  break;
+	case 'H':
+	  *p = 0;
+	  if (strlen(HostName) < l)
+	    {
+	      strcpy(p, HostName);
+	      if (*p)
+		qmflag = 1;
+	    }
+	  p += strlen(p) - 1;
+	  break;
+	case 'F':
+	  p--;
+	  /* small hack */
+	  if ((ev && ev == &D_forecv->c_captev) || (!ev && win && win == D_fore))
+	    qmflag = 1;
+	  break;
+	case '>':
+	  truncpos = p - winmsg_buf;
+	  truncper = num > 100 ? 100 : num;
+	  trunclong = longflg;
+	  p--;
+	  break;
+	case '=':
+	case '<':
+	  *p = ' ';
+	  if (num || zeroflg || plusflg || longflg || (*s != '='))
+	    {
+	      /* expand all pads */
+	      if (minusflg)
+		{
+		  num = (plusflg ? lastpad : padlen) - num;
+		  if (!plusflg && padlen == 0)
+		    num = p - winmsg_buf;
+		  plusflg = 0;
+		}
+	      else if (!zeroflg)
+		{
+		  if (*s != '=' && num == 0 && !plusflg)
+		    num = 100;
+		  if (num > 100)
+		    num = 100;
+		  if (padlen == 0)
+		    num = p - winmsg_buf;
+		  else
+		    num = (padlen - (plusflg ? lastpad : 0)) * num / 100;
+		}
+	      if (num < 0)
+		num = 0;
+	      if (plusflg)
+		num += lastpad;
+	      if (num > MAXSTR - 1)
+		num = MAXSTR - 1;
+	      if (numpad)
+	        p = pad_expand(winmsg_buf, p, numpad, num);
+	      numpad = 0;
+	      if (p - winmsg_buf > num && !longflg)
+		{
+		  int left, trunc;
+
+		  if (truncpos == -1)
+		    {
+		      truncpos = lastpad;
+		      truncper = 0;
+		    }
+		  trunc = lastpad + truncper * (num - lastpad) / 100;
+		  if (trunc > num)
+		    trunc = num;
+		  if (trunc < lastpad)
+		    trunc = lastpad;
+		  left = truncpos - trunc;
+		  if (left > p - winmsg_buf - num)
+		    left = p - winmsg_buf - num;
+		  debug3("truncpos = %d, trunc = %d, left = %d\n", truncpos, trunc, left);
+		  if (left > 0)
+		    {
+		      if (left + lastpad > p - winmsg_buf)
+			left = p - winmsg_buf - lastpad;
+		      if (p - winmsg_buf - lastpad - left > 0)
+		        bcopy(winmsg_buf + lastpad + left, winmsg_buf + lastpad,  p - winmsg_buf - lastpad - left);
+		      p -= left;
+		      r = winmsg_numrend;
+		      while (r && winmsg_rendpos[r - 1] > lastpad)
+			{
+			  r--;
+			  winmsg_rendpos[r] -= left;
+			  if (winmsg_rendpos[r] < lastpad)
+			    winmsg_rendpos[r] = lastpad;
+			}
+		      if (trunclong)
+			{
+			  if (p - winmsg_buf > lastpad)
+			    winmsg_buf[lastpad] = '.';
+			  if (p - winmsg_buf > lastpad + 1)
+			    winmsg_buf[lastpad + 1] = '.';
+			  if (p - winmsg_buf > lastpad + 2)
+			    winmsg_buf[lastpad + 2] = '.';
+			}
+		    }
+		  if (p - winmsg_buf > num)
+		    {
+		      p = winmsg_buf + num;
+		      if (trunclong)
+			{
+			  if (num - 1 >= lastpad)
+			    p[-1] = '.';
+			  if (num - 2 >= lastpad)
+			    p[-2] = '.';
+			  if (num - 3 >= lastpad)
+			    p[-3] = '.';
+			}
+		      r = winmsg_numrend;
+		      while (r && winmsg_rendpos[r - 1] > num)
+			winmsg_rendpos[--r] = num;
+		    }
+		  truncpos = -1;
+		  trunclong = 0;
+		  if (lastpad > p - winmsg_buf)
+		    lastpad = p - winmsg_buf;
+		}
+	      if (*s == '=')
+		{
+		  while (p - winmsg_buf < num)
+		    *p++ = ' ';
+		  lastpad = p - winmsg_buf;
+		  truncpos = -1;
+		  trunclong = 0;
+		}
+	      p--;
+	    }
+	  else if (padlen)
+	    {
+	      *p = 127;		/* internal pad representation */
+	      numpad++;
+	    }
+	  break;
 	case 'n':
 	  s++;
 	  /* FALLTHROUGH */
@@ -2258,6 +2514,12 @@ struct event *ev;
   if (qmpos && !qmflag)
     p = qmpos + 1;
   *p = '\0';
+  if (numpad)
+    {
+      if (padlen > MAXSTR - 1)
+	padlen = MAXSTR - 1;
+      p = pad_expand(winmsg_buf, p, numpad, padlen);
+    }
   if (ev)
     {
       evdeq(ev);		/* just in case */
@@ -2284,7 +2546,7 @@ char *s;
 struct win *win;
 int esc;
 {
-  return MakeWinMsgEv(s, win, esc, (struct event *)0);
+  return MakeWinMsgEv(s, win, esc, 0, (struct event *)0);
 }
 
 int
@@ -2438,9 +2700,13 @@ char *data;
       debug("Backend received interrupt\n");
       /* This approach is rather questionable in a multi-display
        * environment */
-      if (fore)
+      if (fore && displays)
 	{
-	  char ibuf = intrc;
+#if defined(TERMIO) || defined(POSIX)
+	  char ibuf = displays->d_OldMode.tio.c_cc[VINTR];
+#else
+	  char ibuf = displays->d_OldMode.m_tchars.t_intrc;
+#endif
 #ifdef PSEUDOS
 	  write(W_UWP(fore) ? fore->w_pwin->p_ptyfd : fore->w_ptyfd, 
 		&ibuf, 1);
@@ -2485,6 +2751,7 @@ char *data;
 	  /* don't annoy the user with two messages */
 	  if (p->w_monitor == MON_FOUND)
 	    p->w_monitor = MON_DONE;
+          WindowChanged(p, 'f');
 	}
       if (p->w_monitor == MON_FOUND)
 	{
@@ -2504,6 +2771,7 @@ char *data;
 	      Msg(0, "%s", MakeWinMsg(ActivityString, p, '%'));
 	      p->w_monitor = MON_DONE;
 	    }
+          WindowChanged(p, 'f');
 	}
     }
 
